'
'
' ####################
' #####  PROLOG  #####
' ####################
'
' A simple gui designer-like demo to show how to
' create/move/resize child controls in a window.
' Controls can be drawn on window, resized, and moved.
' A right-click context menu can be used to modify
' or delete a control.
'
VERSION	"0.0001"
'
'	IMPORT	"xst"   		' standard library	: required by most programs
	IMPORT  "xst_s.lib"
	IMPORT  "xsx_s.lib"
'	IMPORT  "xsx"				' extended std library
'	IMPORT	"xio"				' console io library
	IMPORT	"gdi32"     ' gdi32.dll
	IMPORT  "user32"    ' user32.dll
	IMPORT  "kernel32"  ' kernel32.dll
	IMPORT  "comctl32"	' comctl32.dll			: common controls library

TYPE CONTROLS
	XLONG	.hwnd
	RECT .rect
	XLONG .type
	XLONG .style
END TYPE

DECLARE FUNCTION Entry            		()
DECLARE FUNCTION WndProc          		(hwnd, msg, wParam, lParam)
DECLARE FUNCTION InitGui          		()
DECLARE FUNCTION RegisterWinClass 		(className$, addrWndProc, icon$, menu$)
DECLARE FUNCTION CreateWindows    		()
DECLARE FUNCTION NewWindow        		(className$, titleBar$, style, x, y, w, h, exStyle)
DECLARE FUNCTION NewChild         		(className$, text$, style, x, y, w, h, parent, id, exStyle)
DECLARE FUNCTION MessageLoop      		()
DECLARE FUNCTION CleanUp          		()
DECLARE FUNCTION CreateControlEx  		(ctype, ic)
DECLARE FUNCTION Initialize       		()
DECLARE FUNCTION ButtonSubClassProc   (hCtl, wMsg, wParam, lParam)
DECLARE FUNCTION EditSubClassProc     (hCtl, wMsg, wParam, lParam)
DECLARE FUNCTION ComboBoxSubClassProc (hCtl, wMsg, wParam, lParam)
DECLARE FUNCTION HitTest              (hCtl, lParam)
DECLARE FUNCTION GetControlInfo       (hCtl, @type, @style, @index)

SHARED CONTROLS hctl[99]
SHARED ghWnd, hInst
SHARED lpButtonProc, lpEditProc, lpComboBoxProc
SHARED fDrawing, ic

' control types
$$CONTROL_STATIC = 2000
$$CONTROL_EDIT = 2100
$$CONTROL_BUTTON = 2200
$$CONTROL_BUTTON_GROUPBOX = 2201
$$CONTROL_COMBOBOX = 2300
$$CONTROL_COMBOBOX_DROPDOWNLIST = 2301

' menu constants
$$IDM_DELETE_CONTROL = 1000
$$IDM_MODIFY_CONTROL = 1001

' style offsets
$$STYLEOFFSET = 1200

FUNCTION Entry ()

	STATIC	entry
'
	IF entry THEN RETURN							' enter once
	entry =  $$TRUE										' enter occured

'	XioCreateConsole (title$, 50)			' create console, if console is not wanted, comment out this line
	InitGui ()												' initialize program and libraries
	IF CreateWindows () THEN QUIT (0)	' create windows and other child controls
	MessageLoop ()										' the main message loop
	CleanUp ()												' unregister all window classes
'	XioFreeConsole ()									' free console

END FUNCTION

FUNCTION WndProc (hWnd, msg, wParam, lParam)

	STATIC x, y
	STATIC fDrag, xStart, yStart
	STATIC POINT pt
	STATIC hdc
	SHARED CONTROLS hctl[]
	SHARED fDrawing, ic
	POINT point, ptclient
	SHARED hChild
	STATIC ctype

	SELECT CASE msg

		CASE $$WM_CREATE : 
			Initialize ()
			ic = 0											' initialize control counter
			ctype = $$CONTROL_BUTTON		' default control type button, 0 = no control
			
		CASE $$WM_CLOSE :
			PostQuitMessage (0)

		CASE $$WM_LBUTTONDOWN :
			pt.x = LOWORD(lParam)
			pt.y = HIWORD(lParam)

			' Start to draw a new object ?
			' This demo makes only 4 controls one-time demo

			IF ic = 4 THEN
				RETURN
			ELSE
				SetCapture (hWnd)
				xStart   =  LOWORD (lParam)
				yStart   =  HIWORD (lParam)
				fDrawing =  $$TRUE
			END IF
			
		CASE $$WM_MOUSEMOVE :

			x = LOWORD (lParam)              ' current X position
			y = HIWORD (lParam)              ' current Y position

		' Are we drawing a new control ?
			IF fDrawing THEN
				x = LOWORD (lParam)            ' current X position
				y = HIWORD (lParam)            ' current Y position
				hdc = GetDC (hWnd)
				SetROP2 (hdc, $$R2_NOTXORPEN)

				' erasing the previous rectangle
				DrawFocusRect (hdc, &hctl[ic].rect)

				' going left?
				IF xStart > x THEN
					hctl[ic].rect.left  = x
					hctl[ic].rect.right = xStart
				ELSE                       
					' going right
					hctl[ic].rect.left  = xStart
					hctl[ic].rect.right = x
				END IF

				' going down?
				IF yStart < y THEN
					hctl[ic].rect.bottom = y
					hctl[ic].rect.top    = yStart
				ELSE                                  
					' going right
					hctl[ic].rect.bottom = yStart
					hctl[ic].rect.top    = y
				END IF

				' show the new position
				DrawFocusRect (hdc, &hctl[ic].rect)
				ReleaseDC (hWnd, hdc)
			END IF

		CASE $$WM_LBUTTONUP:

			x = LOWORD (lParam)              ' current X position
			y = HIWORD (lParam)              ' current Y position

			IF fDrag THEN
				fDrag = $$FALSE
				ReleaseCapture ()
				RETURN DefWindowProcA (hWnd, msg, wParam, lParam)
			END IF

			IF fDrawing THEN
				ReleaseCapture ()

				' going left?
				IF xStart > x THEN
					hctl[ic].rect.left  = x
					hctl[ic].rect.right = xStart
				ELSE    
					' going right
					hctl[ic].rect.left  = xStart
					hctl[ic].rect.right = x
				END IF

				' going down?
				IF yStart < y THEN
					hctl[ic].rect.bottom = y
					hctl[ic].rect.top    = yStart
				ELSE    
					' going right
					hctl[ic].rect.bottom = yStart
					hctl[ic].rect.top    = y
				END IF

				' create a control based on type
				SELECT CASE ic
					CASE 0 : CreateControlEx ($$CONTROL_BUTTON, ic)
					CASE 1 : CreateControlEx ($$CONTROL_BUTTON_GROUPBOX, ic)
					CASE 2 : CreateControlEx ($$CONTROL_EDIT, ic)
					CASE 3 : CreateControlEx ($$CONTROL_COMBOBOX_DROPDOWNLIST, ic)
				END SELECT
				INC ic

				' refresh the main window
				InvalidateRect (hWnd, 0, 1)
				UpdateWindow (hWnd)
				ReleaseCapture ()

				' the drawing is over
				fDrawing = $$FALSE
			END IF
			
		CASE $$WM_RBUTTONDOWN   :
			' get current cursor position for context menu
			GetCursorPos (&point)
			ptclient = point
			ScreenToClient (hWnd, &ptclient)
			
			' see if a child control was clicked on
			hChild = ChildWindowFromPoint (hWnd, ptclient.x, ptclient.y)
			IFZ hChild THEN RETURN
			IF hChild = hWnd THEN RETURN
			
			' create popup menu for child control
			hPopupMenu = CreatePopupMenu ()
			AppendMenuA (hPopupMenu, $$MF_STRING, $$IDM_DELETE_CONTROL, &"Delete Control")

			' create a popup menu to change control style
			' not sure how useful this would be, but shows it can be done 
			GetControlInfo (hChild, @type, 0, 0)
			SELECT CASE type
				CASE $$CONTROL_BUTTON, $$CONTROL_BUTTON_GROUPBOX :
					hSubMenu = CreateMenu()
					AppendMenuA (hPopupMenu, $$MF_STRING | $$MF_POPUP, hSubMenu, &"Modify Control Style")
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_3STATE + $$STYLEOFFSET, &"BS_3STATE")
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_AUTO3STATE + $$STYLEOFFSET, &"BS_AUTO3STATE")
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_AUTOCHECKBOX + $$STYLEOFFSET, &"BS_AUTOCHECKBOX")
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_CHECKBOX + $$STYLEOFFSET, &"BS_CHECKBOX")	
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_DEFPUSHBUTTON + $$STYLEOFFSET, &"BS_DEFPUSHBUTTON")	
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_GROUPBOX + $$STYLEOFFSET, &"BS_GROUPBOX")
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_PUSHBUTTON + $$STYLEOFFSET, &"BS_PUSHBUTTON")
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_RADIOBUTTON + $$STYLEOFFSET, &"BS_RADIOBUTTON")	
					AppendMenuA (hSubMenu, $$MF_STRING, $$BS_BITMAP + $$STYLEOFFSET, &"BS_BITMAP")	
			END SELECT
			
			TrackPopupMenu (hPopupMenu, 0, point.x, point.y, 0, hWnd, 0)
			DestroyMenu (hSubMenu)
			DestroyMenu (hPopupMenu)
			UpdateWindow (hWnd)
			
		CASE $$WM_COMMAND :
			id         = LOWORD(wParam)
			hwndCtl    = lParam
			notifyCode = HIWORD(wParam)

			' modify style of button
			IF id >= $$STYLEOFFSET THEN 
				style = id - $$STYLEOFFSET													' get new style
				GetControlInfo (hChild, @type, @oldStyle, @index)		' get control's current style and index
				currentStyle = GetWindowLongA (hChild, $$GWL_STYLE)	' get current window style
				newStyle = currentStyle & ~oldStyle | style					' create new style
				SetWindowLongA (hChild, $$GWL_STYLE, newStyle)			' set new style
				hctl[index].style = style														' update style in chtl[] array
				UpdateWindow (hChild)
				InvalidateRect (hWnd, 0, 1) 
				RETURN
			END IF
			 
			SELECT CASE id
				
				' delete a control
				CASE $$IDM_DELETE_CONTROL :
					ret = MessageBoxA (hWnd, &"Do you wish to delete this control?", &"Delete Control", $$MB_ICONEXCLAMATION | $$MB_YESNOCANCEL)
					SELECT CASE ret
						CASE $$IDCANCEL, $$IDNO :
						CASE $$IDYES : 
							DestroyWindow (hChild)
							UpdateWindow (hWnd)
							FOR i = 0 TO UBOUND(hctl[])
								IF hctl[i].hwnd = hChild THEN
									hctl[i].hwnd = 0
									' do some other things if necessary to correctly update hctl[] array
									hChild = 0
									EXIT FOR
								END IF
							NEXT
					END SELECT	

'				CASE ELSE :
'					text$ = "You clicked on popup menu item " + STRING$(id)
'					MessageBoxA (hWnd, &text$, &"Popup Menu Test", 0)
			END SELECT

		CASE $$WM_DESTROY :
			PostQuitMessage(0)

		CASE ELSE :
			RETURN DefWindowProcA (hWnd, msg, wParam, lParam)

	END SELECT

END FUNCTION

FUNCTION InitGui ()

	SHARED hInst
	hInst = GetModuleHandleA (0) 	' get current instance handle
	IFZ hInst THEN QUIT (0)
	InitCommonControls ()					' initialize comctl32.dll library

END FUNCTION

FUNCTION RegisterWinClass (className$, addrWndProc, icon$, menu$)

	SHARED hInst
	WNDCLASS wc

	wc.style           = $$CS_HREDRAW | $$CS_VREDRAW | $$CS_OWNDC
	wc.lpfnWndProc     = addrWndProc
	wc.cbClsExtra      = 0
	wc.cbWndExtra      = 0
	wc.hInstance       = hInst
	wc.hIcon           = LoadIconA (hInst, &icon$)
	wc.hCursor         = LoadCursorA (0, $$IDC_ARROW)
	wc.hbrBackground   = $$COLOR_BTNFACE + 1
	wc.lpszMenuName    = &menu$
	wc.lpszClassName   = &className$

	IFZ RegisterClassA (&wc) THEN RETURN ($$TRUE)

END FUNCTION

FUNCTION CreateWindows ()

	SHARED ghWnd

' register window class
	className$  = "designer"
	addrWndProc = &WndProc()
	icon$ 			= "scrabble"
	menu$ 			= ""
	IF RegisterWinClass (@className$, addrWndProc, @icon$, @menu$) THEN RETURN ($$TRUE)

' create main window
	titleBar$ = "Add Controls + Move + Resize"
	style 		= $$WS_OVERLAPPEDWINDOW
	w 				= 480
	h 				= 360
	exStyle		= 0
	ghWnd = NewWindow (@className$, @titleBar$, style, x, y, w, h, exStyle)
	IFZ ghWnd THEN RETURN ($$TRUE)

	XstCenterWindow (ghWnd)							' center window position
	ShowWindow (ghWnd, $$SW_SHOWNORMAL)	' show window

END FUNCTION

FUNCTION NewWindow (className$, titleBar$, style, x, y, w, h, exStyle)

	SHARED hInst

	RETURN CreateWindowExA (exStyle, &className$, &titleBar$, style, x, y, w, h, 0, 0, hInst, 0)

END FUNCTION

FUNCTION NewChild (className$, text$, style, x, y, w, h, parent, id, exStyle)

	SHARED hInst

	style = style | $$WS_CHILD | $$WS_VISIBLE
	RETURN CreateWindowExA (exStyle, &className$, &text$, style, x, y, w, h, parent, id, hInst, 0)

END FUNCTION

FUNCTION MessageLoop ()

	MSG msg

' main message loop

	IF LIBRARY(0) THEN RETURN								' main program executes message loop

	DO																			' the message loop
		ret = GetMessageA (&msg, NULL, 0, 0)	' retrieve next message from queue

		SELECT CASE ret
			CASE  0 : RETURN msg.wParam					' WM_QUIT message
			CASE -1 : RETURN $$TRUE							' error
			CASE ELSE:
				hwnd = GetActiveWindow ()
				IF (!IsWindow (hwnd)) || (!IsDialogMessageA (hwnd, &msg)) THEN		' send only non-dialog messages
  				TranslateMessage (&msg)					' translate virtual-key messages into character messages
  				DispatchMessageA (&msg)					' send message to window callback function WndProc()
				END IF
		END SELECT
	LOOP

END FUNCTION
FUNCTION CleanUp ()

	SHARED hInst, className$

	UnregisterClassA (&className$, hInst)

END FUNCTION

FUNCTION CreateControlEx (ctype, ic)

	SHARED hInst, ghWnd
	SHARED lpButtonProc, lpEditProc, lpComboBoxProc
	SHARED CONTROLS hctl[]
	
	ws = $$WS_CHILD OR $$WS_VISIBLE OR $$WS_CLIPSIBLINGS 

	SELECT CASE ctype
	
		CASE 0          
			' Reserved for no control selected.
	
		CASE $$CONTROL_BUTTON	
			style = $$BS_PUSHBUTTON 
			hWndNew  = CreateWindowExA (0, &"BUTTON", &"B&utton 1", _
			style OR ws, _
			hctl[ic].rect.left, _
			hctl[ic].rect.top, _
			hctl[ic].rect.right  - hctl[ic].rect.left, _
			hctl[ic].rect.bottom - hctl[ic].rect.top, _
			ghWnd, 0, hInst, 0)
		
			lpButtonProc = SetWindowLongA (hWndNew, $$GWL_WNDPROC, &ButtonSubClassProc())
	
		CASE $$CONTROL_BUTTON_GROUPBOX	
			style = $$BS_GROUPBOX 		
			hWndNew  = CreateWindowExA (0, &"BUTTON", &"GroupBox 1", _
			style OR ws, _
			hctl[ic].rect.left, _
			hctl[ic].rect.top, _
			hctl[ic].rect.right  - hctl[ic].rect.left, _
			hctl[ic].rect.bottom - hctl[ic].rect.top, _
			ghWnd, 0, hInst,  0)
		
			lpButtonProc = SetWindowLongA (hWndNew, $$GWL_WNDPROC, &ButtonSubClassProc())
	
		CASE $$CONTROL_EDIT	
			style = $$WS_BORDER OR $$WS_VSCROLL 					
			hWndNew  = CreateWindowExA (0, &"EDIT", &"Edit 1", _
			style OR ws, _
			hctl[ic].rect.left, _
			hctl[ic].rect.top, _
			hctl[ic].rect.right  - hctl [ic].rect.left, _
			hctl[ic].rect.bottom - hctl [ic].rect.top, _
			ghWnd, 0, hInst, 0)
		
			lpEditProc = SetWindowLongA (hWndNew, $$GWL_WNDPROC, &EditSubClassProc())
	
		CASE $$CONTROL_COMBOBOX_DROPDOWNLIST
			style = $$WS_BORDER OR $$CBS_DROPDOWNLIST OR $$CBS_NOINTEGRALHEIGHT					
			hWndNew  = CreateWindowExA (0, &"COMBOBOX", &"ComboBox 1", _
			style | ws, _
			hctl[ic].rect.left, _
			hctl[ic].rect.top, _
			hctl[ic].rect.right  - hctl[ic].rect.left, _
			hctl[ic].rect.bottom - hctl[ic].rect.top, _
			ghWnd, 0, hInst, 0)
		
			s$ = "ComboBox 1"
		
			SendMessageA (hWndNew, $$CB_ADDSTRING, 0, &s$)
			SendMessageA (hWndNew, $$CB_SETCURSEL, 0, 0)
		
			lpComboBoxProc = SetWindowLongA (hWndNew, $$GWL_WNDPROC, &ComboBoxSubClassProc())
	
	END SELECT
	
	'  save handle and rect structure
	hctl[ic].hwnd        = hWndNew
	hctl[ic].type        = ctype
	hctl[ic].style			 = style
	hctl[ic].rect.left   = hctl[ic].rect.left
	hctl[ic].rect.top    = hctl[ic].rect.top
	hctl[ic].rect.right  = hctl[ic].rect.right  - hctl[ic].rect.left
	hctl[ic].rect.bottom = hctl[ic].rect.bottom - hctl[ic].rect.top

END FUNCTION

FUNCTION Initialize ()

MyMsg$ = _
				"          A simple control designer demo."                 _
			+ "\r\n" + "\r\n"                                             _
			+ "After the form appears, left click and drag anywhere "     _
			+ "\r\n"                                              				_
			+ "on the form and then release to create up to 4 controls."  _
			+ "\r\n"																											_
			+ "Right click on any control to display a popup menu."				_
			+ "\r\n" + "\r\n"                                             _
			+ "Based on code by Jules Marchildon." 

	MessageBoxA (NULL, &MyMsg$, _
  &"Control Designer", $$MB_OK OR $$MB_ICONINFORMATION OR $$MB_SYSTEMMODAL)

	SetCursor (LoadCursorA (NULL, $$IDC_ARROW))

END FUNCTION

FUNCTION ButtonSubClassProc (hCtl, wMsg, wParam, lParam)

	SHARED lpButtonProc

	SELECT CASE wMsg
		CASE $$WM_RBUTTONDOWN
		CASE $$WM_LBUTTONDBLCLK 		' needs CS_DBLCLKS style
		CASE $$WM_NCHITTEST
			RETURN HitTest (hCtl, lParam)
		CASE $$WM_SIZING   					' Gridize while sizing
		CASE $$WM_MOVING   					' Gridize while moving
		CASE $$WM_NCRBUTTONDOWN
			ScreenToClient (GetParent (hCtl), &lParam)
			SendMessageA (GetParent (hCtl), $$WM_RBUTTONDOWN, wParam, lParam)
	END SELECT
	RETURN CallWindowProcA (lpButtonProc, hCtl, wMsg, wParam, lParam)
END FUNCTION

FUNCTION EditSubClassProc (hCtl, wMsg, wParam, lParam)

	SHARED lpEditProc
	
	SELECT CASE wMsg
		CASE $$WM_RBUTTONDOWN
		CASE $$WM_LBUTTONDBLCLK    	'needs CS_DBLCLKS style
		CASE $$WM_NCHITTEST
			RETURN HitTest (hCtl, lParam)
		CASE $$WM_SIZING   					' Gridize while sizing
		CASE $$WM_MOVING   					' Gridize while moving
		CASE $$WM_NCRBUTTONDOWN
			ScreenToClient (GetParent (hCtl), &lParam)
			SendMessageA (GetParent (hCtl), $$WM_RBUTTONDOWN, wParam, lParam)
	END SELECT
	RETURN CallWindowProcA (lpEditProc, hCtl, wMsg, wParam, lParam)

END FUNCTION

FUNCTION ComboBoxSubClassProc (hCtl, wMsg, wParam, lParam)

	SHARED lpComboBoxProc
	
	SELECT CASE wMsg
		CASE $$WM_RBUTTONDOWN
		CASE $$WM_LBUTTONDBLCLK     ' needs CS_DBLCLKS style
		CASE $$WM_NCHITTEST
			RETURN HitTest (hCtl, lParam)
		CASE $$WM_SIZING   					' Gridize while sizing
		CASE $$WM_MOVING   					' Gridize while moving
		CASE $$WM_NCRBUTTONDOWN
			ScreenToClient (GetParent (hCtl), &lParam)
			SendMessageA (GetParent (hCtl), $$WM_RBUTTONDOWN, wParam, lParam)
	END SELECT
	RETURN CallWindowProcA (lpComboBoxProc, hCtl, wMsg, wParam, lParam)

END FUNCTION

FUNCTION HitTest (hCtl, lParam)

	POINT pt
	RECT rc 
	
	pt.x = LOWORD (lParam)
	pt.y = HIWORD (lParam)
	
	ScreenToClient (hCtl, &pt)
	GetWindowRect  (hCtl, &rc)
	
	MapWindowPoints ($$HWND_DESKTOP, ghWnd, &rc, 2)

	SELECT CASE TRUE
		CASE pt.y < 4 && pt.x < 4
			RETURN $$HTTOPLEFT
		CASE pt.y < 4 && pt.x >= (rc.right-rc.left-4)
			RETURN $$HTTOPRIGHT
		CASE pt.y >=(rc.bottom-rc.top-4) && pt.x >= (rc.right-rc.left-4)
			RETURN $$HTBOTTOMRIGHT
		CASE pt.x < 4 && pt.y >= (rc.bottom -rc.top - 4)
			RETURN $$HTBOTTOMLEFT
		CASE pt.y < 4
			RETURN $$HTTOP
		CASE pt.x < 4
			RETURN $$HTLEFT
		CASE pt.x >= (rc.right -rc.left - 4)
			RETURN $$HTRIGHT
		CASE pt.y >= (rc.bottom -rc.top - 4)
			RETURN $$HTBOTTOM
		CASE ELSE
			RETURN $$HTCAPTION
	END SELECT

END FUNCTION

FUNCTION GetControlInfo (hCtl, @type, @style, @index)

	SHARED CONTROLS hctl[]
	
	upp = UBOUND(hctl[])
	FOR i = 0 TO upp
		IF hctl[i].hwnd = hCtl THEN
			type =  hctl[i].type
			style = hctl[i].style
			index = i
		END IF
	NEXT i
	
END FUNCTION

END PROGRAM
