'
' ####################
' #####  PROLOG  #####
' ####################
'
' A smtp mail library using Xin
'	For information on the SMTP protocol: http://www.faqs.org/rfcs/rfc821.html
'
PROGRAM "xinsmtp"
VERSION "0.0001"
CONSOLE
'
	IMPORT "xst"				' Standard library : required by most programs
	IMPORT "xin"
	
'	IMPORT "xst_s.lib"
' IMPORT "xin_s.lib"

	IMPORT "kernel32"	  ' kernel32.dll

EXPORT
DECLARE FUNCTION XinSmtp ()
DECLARE FUNCTION SmtpConnect (serverName$, @socket)
DECLARE FUNCTION SmtpSendCmd (cmd$, message$, socket)
DECLARE FUNCTION SmtpReadResponse (socket, @response$)
DECLARE FUNCTION SmtpEvalResponse (response$)
DECLARE FUNCTION SmtpQuit (socket)
DECLARE FUNCTION SmtpSetTimeout (usec)
DECLARE FUNCTION SmtpMailFrom (from$, socket)
DECLARE FUNCTION SmtpRcptTo (@rcpts$[], socket)
DECLARE FUNCTION SmtpSendMail (from$, fromName$, @rcpts$[], subject$, body$, server$)
DECLARE FUNCTION SmtpWrite (socket, text$)
DECLARE FUNCTION SmtpGetTimeStamp (@date$)
DECLARE FUNCTION SmtpEcho (fEcho)
END EXPORT
'
'
' ######################
' #####  Entry ()  #####
' ######################
'
FUNCTION XinSmtp ()

	SHARED echo
	SHARED timeout
	
	Xin ()
	timeout = 1000
	IF LIBRARY(0) THEN RETURN	

	DIM rcpts$[0]	
	
	server$ 		= INLINE$ ("server name>") 			' "mail.aol.com"
	from$ 			= INLINE$ ("from address>") 		' "mac.intosh@aol.com"
	fromName$ 	= INLINE$ ("from name>") 				' "Mac Intosh"
	rcpts$[0] 	= INLINE$ ("recipient address>")' "joe.smith@yahoo.com"
	subject$ 		= INLINE$ ("subject>")  				' "Re: Your Birthday"
	body$ 			= INLINE$ ("message>")  				' "Happy BD"	
	
	IFZ SmtpSendMail (from$, fromName$, @rcpts$[], subject$, body$, server$) THEN
		? "message sent"
	ELSE
	  ? "error sending message"
	END IF

  INLINE$ ("Press any key to quit >")

END FUNCTION

FUNCTION SmtpConnect (serverName$, @socket)
' SmtpConnect - connect to remote server

	SHARED echo
	SHARED timeout
	HOST  host
	serverPort = 25
	block = timeout
	
	XinHostNumberToInfo (0, @host)
	localHost = host.address

	FOR i = 0 TO 7
		IF host.addresses[i] > 0 THEN
			localHost = host.addresses[i]
		ELSE
			EXIT FOR
		END IF
	NEXT i

	XinAddressNumberToString (localHost, @addr$)
	
	IF XinSocketOpen (@socket, 0, 0, 0) THEN
		? "SmtpConnect: XinSocketOpen error"
		XinSocketClose (-1)
		RETURN $$TRUE
	END IF
	
	port = 0
	address$$ = 0
	IF XinSocketBind (socket, block, @address$$, @port) THEN
		? "SmtpConnect: XinSocketBind error"
		XinSocketClose (-1)
		RETURN $$TRUE
	END IF

	XinHostNameToInfo (serverName$, @host)
	XinAddressNumberToString (host.address, @addr$)

	error = XinSocketConnectRequest (socket, 0, host.address, serverPort)
	IF error THEN
		? "SmtpConnect: XinSocketConnectRequest error"
		XinSocketClose (-1)
		RETURN $$TRUE
	END IF

	DO UNTIL connected
		error = XinSocketConnectStatus (socket, block, @connected)
		IF error THEN
			? "SmtpConnect: XinSocketConnectStatus error"
			XinSocketClose (-1)
			RETURN $$TRUE
		END IF
	LOOP
	
	IF SmtpSendCmd ("HELO", addr$, socket) THEN RETURN ($$TRUE)
	IF SmtpReadResponse (socket, @response$) THEN RETURN ($$TRUE)
	
	RETURN SmtpEvalResponse (response$)
	
END FUNCTION

FUNCTION SmtpSendCmd (cmd$, message$, socket)
' SmtpSendCmd - send a smtp command

	SHARED timeout

	IFZ cmd$ THEN RETURN $$TRUE
	
	block = timeout
	
	IF message$ THEN
		cmd$ = cmd$ + " " + message$
	ELSE
		cmd$ = cmd$
	ENDIF
	
	RETURN SmtpWrite (socket, cmd$)

END FUNCTION

FUNCTION SmtpReadResponse (socket, @response$)
' SmtpReadResponse - read a response from the server

	SHARED timeout

	response$ = NULL$(256)
	block = timeout
	bytes = 0
	error = 0
	
	DO UNTIL error
		error = XinSocketRead (socket, block, &response$, 256, 0, @bytes)
		IF error THEN 
			? "SmtpReadResponse: XinSocketRead error"
			RETURN $$TRUE
		END IF
	LOOP UNTIL bytes

	response$ = TRIM$(CSIZE$(response$))

END FUNCTION

FUNCTION SmtpEvalResponse (response$)
' SmtpEvalResponse - evaluate server response

	IFZ response$ THEN RETURN ($$TRUE)
	
	IF LEN(response$) < 3 THEN
		? "Unknown server response"
		RETURN ($$TRUE)
	END IF
	
	SELECT CASE LEFT$(response$, 3)
	
		CASE "421", "432", "450", "451", "452", "454", "500", "501", _
         "502", "503", "504", "530", "534", "535", "538", "550", _
         "552", "553", "554" : 
			? "Fatal error"
			? response$
			RETURN ($$TRUE)
			
		CASE "211", "214", "251", "252", "551" :
			? "Server error"
			? response$
			RETURN
			
		CASE "220", "221", "235", "250", "334", "354" :  
			? "Ok"
			? response$
		
		CASE ELSE
			? "Unknown error"
			? response$
			RETURN ($$TRUE)
	
	END SELECT

END FUNCTION

FUNCTION SmtpQuit (socket)
' SmtpQuit - disconnect from server

	SHARED echo
	
	cmd$ = "QUIT"
	IF echo THEN ? cmd$

	IF SmtpSendCmd (cmd$, "", socket) THEN RETURN ($$TRUE)
	
	ret = SmtpReadResponse (socket, @response$)
	
	XinSocketClose (-1)
	
	RETURN ret

END FUNCTION

FUNCTION SmtpSetTimeout (usec)
' SmtpSetTimeout - set blocking timeout in usec
'	timeout < 1			: timout = 0				' do not block
'	timeout < 10		: timeout = 10			' min block = 10us
'	timeout > 20000	: timeout = 20000		' max block = 20ms
	SHARED timeout
	
	timeout = usec

END FUNCTION

FUNCTION SmtpMailFrom (from$, socket)
' SmtpMailFrom - sends the MAIL FROM command

	SHARED echo
	
	IFZ from$ THEN RETURN ($$TRUE)
	
	from$ = "<" + from$ + ">"
	cmd$ = "MAIL FROM:"
	IF echo THEN ? cmd$ + " " + from$

	IF SmtpSendCmd (cmd$, from$, socket) THEN RETURN ($$TRUE)
	IF SmtpReadResponse (socket, @response$) THEN RETURN ($$TRUE)
	RETURN SmtpEvalResponse (response$)

END FUNCTION

FUNCTION SmtpRcptTo (@rcpts$[], socket)
' SmtpRcptTo - sends the RCPT TO command

	SHARED echo
	
	IFZ rcpts$[] THEN RETURN ($$TRUE)
	
	upp = UBOUND (rcpts$[])
	
	FOR i = 0 TO upp
	
		to$ = "<" + rcpts$[i] + ">"
		cmd$ = "RCPT TO:"
		IF echo THEN ? cmd$ + " " + to$
	
		IF SmtpSendCmd (cmd$, to$, socket) THEN RETURN ($$TRUE)
		IF SmtpReadResponse (socket, @response$) THEN RETURN ($$TRUE)
		IF SmtpEvalResponse (response$) THEN RETURN ($$TRUE)
		
		INC valid
	NEXT i
	
	IFZ valid THEN
		? "No valid recipient was given"
		RETURN ($$TRUE)
	END IF

END FUNCTION

FUNCTION SmtpSendMail (from$, fromName$, @rcpts$[], subject$, body$, server$)
' SmtpSendMail - connect to server and sends mail using DATA command

	SHARED echo

	IFZ rcpts$[] THEN RETURN ($$TRUE)
	
	IF SmtpConnect (server$, @socket) THEN RETURN ($$TRUE)
	IF SmtpMailFrom (from$, socket) THEN GOTO quit
	IF SmtpRcptTo (@rcpts$[], socket) THEN GOTO quit
	
	' send the DATA command
	
	cmd$ = "DATA"
	
	IF echo THEN ? cmd$
	IF SmtpSendCmd (cmd$, "", socket) THEN GOTO quit
	IF SmtpReadResponse (socket, @response$) THEN GOTO quit
	IF SmtpEvalResponse (response$) THEN GOTO quit

	SmtpGetTimeStamp (@date$)
	IF SmtpWrite (socket, "Date: " + date$) THEN GOTO quit
	to$ = rcpts$[0]
	IF SmtpWrite (socket, "To: <" + to$ + ">") THEN GOTO quit
	IF SmtpWrite (socket, "From: " + fromName$) THEN GOTO quit
	IF SmtpWrite (socket, "Sender: " + from$) THEN GOTO quit
	IF SmtpWrite (socket, "Reply-To: <" + from$ + ">") GOTO quit
	IF SmtpWrite (socket, "Subject: " + subject$) THEN GOTO quit
	IF SmtpWrite (socket, body$) THEN GOTO quit
	IF SmtpWrite (socket, ".") THEN GOTO quit
	SmtpQuit (socket)
	RETURN
	
quit:	
	SmtpQuit (socket)
	RETURN ($$TRUE)


END FUNCTION

FUNCTION SmtpWrite (socket, text$)
' SmtpWrite - write text to socket

	IFZ text$ THEN RETURN ($$TRUE)
	IFZ socket THEN RETURN ($$TRUE)
	
	text$ = text$ + "\r\n"

	writebytes = LEN(text$)
	bytes = 0
	error = XinSocketWrite (socket, block, &text$, writebytes, 0, @bytes)
	IF error THEN
		? "SmtpWrite: XinSocketWrite error"
		RETURN $$TRUE
	END IF

END FUNCTION

FUNCTION SmtpGetTimeStamp (date$)

' SmtpGetTimeStamp - get time and date stamp

SYSTEMTIME systime

	GetSystemTime (&systime)

	format$ = "ddd',' dd MMM yyyy"
	d$ = NULL$(24)
	err = GetDateFormatA (NULL, NULL, &systime, &format$, &d$,	LEN(d$))
	d$ = CSIZE$(d$)

	format$ = "HH':'mm':'ss tt"
	t$ = NULL$(24)
	GetTimeFormatA (NULL, $$TIME_FORCE24HOURFORMAT, &systime, &format$, &t$, LEN(t$))
	t$ = CSIZE$(t$)
	
	date$ = d$ + " " + t$

END FUNCTION

FUNCTION SmtpEcho (fEcho)
' SmtpEcho - if fEcho is $$TRUE, print commands sent

SHARED echo
	echo = fEcho

END FUNCTION

END PROGRAM